#include <cstring>
#include <cstdlib>
#include <cstdio>

#include "myspell.hxx"

/*
 *  MIspell -- (Magyar) MySpell's Ispell Interface
 *  2002 (c) Nmeth Lszl <nemethl@gyorsposta.hu>
 *  License: GNU GPL, version 2 or later
 */

//#define WINDOWS // Windows environment

#define MISPELL_HEADING "@(#) International Ispell Version 3.1.20 - MIspell "
#define MISPELL_VERSION "0.7.2"
#define LOG 0

// interactive interface
#define CURSES

#define BUFSIZ 8192

#ifdef WINDOWS
#undef CURSES
#define LIBDIR "C:\\MySpell\\"
#define HOME "C:\\MySpell\\"
#define DICBASENAME "ispell_"
#define LOGFILE "C:\\MySpell\\log"
#define PATHSEP '\\'
#else
#include <unistd.h>
#ifdef CURSES
#include <curses.h>
#include <locale.h>
#include <libintl.h>
#else
#define gettext
#endif
#define LIBDIR "/usr/lib/myspell/"
#define HOME getenv("HOME")
#define DICBASENAME ".ispell_"
#define LOGFILE "/tmp/mispell.log"
#define PATHSEP '/'
#endif

extern char * mystrdup(const char * s);

//using namespace std;

struct wordlist {
    char * word;
    wordlist * next;
};

// global variables

char * wordchars=NULL;
int wordchars_len;
char * dicname;
int modified;

void log(char * message)
{
    if (LOG) {
        FILE *f = fopen(LOGFILE,"a");
	if (f) {
    	    fprintf(f,"%s\n",message);
    	    fclose(f);
	} else {
	    fprintf(stderr,"Logfile...");
	}
    }
}


int putdic(char * word, MySpell * pMS)
{
    char * w;
    char c;

    pMS->spell("PuT");
    if (((w=strstr(word+1,"PuT"))!=NULL)) {
	c=*w;
	*w='\0';
	pMS->spell(word);
	*w=c;
	return pMS->spell(w+3);				
    } else if (((w=strstr(word+1,"//"))!=NULL)) {
	c=*w;
	*w='\0';
	pMS->spell(word);
	*w=c;
	return pMS->spell(w+2);
    } else {
	pMS->spell(word);
	return pMS->spell(word);
    }
}

void load_privdic(char * filename, MySpell * pMS) 
{
    char buf[BUFSIZ];
	
    FILE *dic = fopen(filename,"r");
    if (dic) {
	while(fgets(buf,BUFSIZ,dic)) {
	    *(buf + strlen(buf) - 1) = '\0';
		putdic(buf,pMS);
	    }
    }
}

int exist(char * filename)
{
    FILE *f = fopen(filename,"r");
    if (f) {
	fclose(f);
    	    return 1;
    }
    return 0;
}

int save_privdic(char * filename, char * filename2, wordlist * w)
{
    wordlist * r;
    FILE *dic = fopen(filename,"r");
    if (dic) {
	fclose(dic);
	dic = fopen(filename,"a");
    } else {
	dic = fopen(filename2,"a");	
    }
    if (! dic) return 0;
    while (w != NULL) {
	fprintf(dic,"%s\n",w->word);
	log(w->word);log("\n");
	r = w;
	free(w->word);
	w = w->next;
	free(r);
    }
    fclose(dic);
    return 1;
}

char * basename(char * s, char c) {
    char * p = s + strlen(s);
    while ((*p != c) && (p != s)) p--;
    if (*p == c) p++;
    return p;
}

void pipe_interface(MySpell * pMS) {
char buf[BUFSIZ];
char * buf2;
wordlist * dicwords = NULL;
int TeXMODE=0;
int k;
int dp;
char * token;
int pos;
char prevnochar='\0';

fprintf(stdout,gettext(MISPELL_HEADING));
fprintf(stdout,MISPELL_VERSION);
fprintf(stdout,"\n");
fflush(stdout);

while(fgets(buf,BUFSIZ,stdin)) {
    log("BE:");
    log(buf);
    k = strlen(buf);
    *(buf + k - 1) = '\0';
    pos=0;

      // execute commands
    switch (buf[0]) {
    case '!': { log("felkiltjel"); break; }
    case '+': { TeXMODE=1; break; }
    case '-': { TeXMODE=0; break; }
    case '@': { putdic(buf+1,pMS); break; }
    case '*': {
        struct wordlist* i = 
	    (struct wordlist *) malloc (sizeof(struct wordlist));
    	i->word = mystrdup(buf+1);
	i->next = dicwords;
	dicwords = i;
        putdic(buf+1,pMS);
        break;
    }
    case '#': {
	strcpy(buf,HOME);
#ifndef WINDOWS
	strcat(buf,"/");
#endif
	buf2 = buf+strlen(buf);
	strcat(buf,DICBASENAME);
	strcat(buf,basename(dicname,PATHSEP));
	if (save_privdic(buf2, buf, dicwords)) {
	    dicwords=NULL;
	}
	break;
    }
    case '^': {
	pos=1;
	// XXX Emacs ispell.el has a hidden mistake:
	// after the replacings often checks a letter
	// we ignore these checks
	if (k==3) {
    	    fprintf(stdout,"\n");
	    fflush(stdout);
	    continue;
	}
	// no break
    }
    default: {
	// data line (not command)
	token = buf+pos;
	for (int i=pos; i<k; i++) {
	    if (!TESTAFF(wordchars, buf[i], wordchars_len)) {
		// first character of line or TeX definition
		if ((buf+i-token==0) || ((TeXMODE==1) && (prevnochar=='\\'))) {
		    // previous non character
		    prevnochar = buf[i];
		    token = buf+i+1;
		    continue;
		}
		prevnochar = buf[i];
		buf[i] = '\0';
		dp = pMS->spell(token);
		if (dp) {
		    if (pMS->get_prevcompound()) {
			fprintf(stdout,"-\n");
		    } else if (pMS->get_prevroot()) {
		    	fprintf(stdout,"+ %s\n",pMS->get_prevroot());
		    } else {
			fprintf(stdout,"*\n");
		    }
		    fflush(stdout);
		} else {
		    int nt = strlen(token);
		    while ((nt != 1) && (*(token+nt-1)=='.')) { // mystake...
			nt--;
			*(token+nt)='\0';
		    }
		    char ** wlst;
		    int ns = pMS->suggest(&wlst,token);
		    if (ns==0) {
		    	fprintf(stdout,"# %s %d",token,token-buf);
			//mistake(line,strlen(token),token-buf);
		    } else {
			fprintf(stdout,"& %s %d %d: %s",token,ns,token-buf,wlst[0]);
			//mistake(line,strlen(token),token-buf);
			//fprintf(stderr,"0 %s  ",wlst[0]);
		    }
		    for (int j=1; j < ns; j++) {
			fprintf(stdout,", %s",wlst[j]);
			//fprintf(stderr,"%d %s  ",j,wlst[j]);
		    	free(wlst[j]);
		    }
		    fprintf(stdout,"\n");
		    fflush(stdout);
		    free(wlst);
		}
		token = buf+i+1; // lower limit
	    }
	} // for
	fprintf(stdout,"\n");
	fflush(stdout);		
    } // default
} // switch
} // while

} // pipe_interface

#ifndef WINDOWS
#ifdef CURSES

void dialogscreen(char * token, char * filename, int len, int begin, char ** wlst,
	int ns, char * newline, int * rel, char * pre2, char * pre) {
    clear();
    printw(gettext("\t%s\t\tFile: %s\n\n"), token,filename);
    if (pre2) printw("%s\n",pre2);
    if (pre) printw("%s\n",pre);
    for (int i=0; i<begin+(*rel); i++) printw("%c",newline[i]);

    attron(A_REVERSE);    
    printw("%s", token);
    attroff(A_REVERSE);
    
    printw("%s\n\n", newline+(*rel)+begin+len);
    
    for (int i=0; i<ns; i++) {
	if ((ns>10) && (i<10)) {
	    printw(" 0%d: %s\n",i,wlst[i]);
	} else {
	    printw(" %d: %s\n",i,wlst[i]);
	}
    }

    int x, y;
    getmaxyx(stdscr,y,x);
    mvprintw(y-3, 1, "%s\n", 
	gettext("\n[SP] <number> R)epl A)ccept I)nsert L)ookup U)ncap Q)uit e(X)it or ? for help\n"));
}


 // for terminal interface
int dialog(MySpell * pMS, char * token, char * filename, char * line, int len,
	int begin, char ** wlst, int ns, char * newline, 
	int * rel, char * pre2, char * pre) {
    char buf[BUFSIZ];
    char * buf2;
    wordlist * dicwords = NULL;
    int c;

    dialogscreen(token, filename, len, begin, wlst, ns, newline, rel, pre2, pre);

    char firstletter='\0';
    
    while (c=getch()) {
    switch (c) {
	case '0':
	case '1': if ((firstletter=='\0') && (ns>10)) {
	    firstletter=c;
	    break;
	}
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9': {
	    modified=1;
	    if ((firstletter!='\0') && (firstletter=='1')) {
		c += 10;
	    }
	    c -= '0';
	    if (c>=ns) break;
	    strcpy(newline+begin+(*rel),wlst[c]);
	    strcat(newline,line+begin+len);
	    *rel+=strlen(wlst[c])-len;
	    ///printw("%d, %s\n",c,newline);
	    goto ki;
	}
	case ' ': { 
	    goto ki;
	}
	case '?': {
	    clear();
printw(gettext("Whenever a word is found that is not in the dictionary\n"));
printw(gettext("it is printed on the first line of the screen.  If the dictionary\n"));
printw(gettext("contains any similar words, they are listed with a number\n"));
printw(gettext("next to each one.  You have the option of replacing the word\n"));
printw(gettext("completely, or choosing one of the suggested words.\n"));
printw(gettext("\nCommands are:\n\n"));
printw(gettext("R	Replace the misspelled word completely.\n"));
printw(gettext("Space	Accept the word this time only.\n"));
printw(gettext("A	Accept the word for the rest of this session.\n"));
printw(gettext("I	Accept the word, and put it in your private dictionary.\n"));
printw(gettext("U	Accept and add lowercase version to private dictionary.\n"));
printw(gettext("O	Ask two words. First word will accepted with 2nd's affixes too.\n"));
printw(gettext("0-n	Replace with one of the suggested words.\n"));
printw(gettext("L	Look up words in system dictionary.\n"));
printw(gettext("X	Write the rest of this file, ignoring misspellings, and start next file.\n"));
printw(gettext("Q	Quit immediately. Asks for confirmation. Leaves file unchanged.\n"));
printw(gettext("!	Shell escape.\n"));
printw(gettext("^L	Redraw screen\n"));
printw(gettext("^Z	Suspend program.\n"));
printw(gettext("?	Show this help screen.\n"));
printw(gettext("\n-- Type space to continue -- \n"));
	    while (getch()!=' ');
	}
	case 12: {
    	    dialogscreen(token, filename, len, begin, wlst, ns, newline, rel, pre2, pre);
	    break;
	}
	default: {
	    if (c==(gettext("r"))[0]) {
		modified=1;
		echo();
		printw(gettext("Replace with: "));
		char i[BUFSIZ];
		getnstr(i,BUFSIZ);
		///fprintf(stderr,"X:%s\n",i);
		strcpy(newline+begin+(*rel),i);
		strcat(newline,line+begin+len);
		// *rel+=strlen(i)-len;
		noecho();
		return 2; // replace
	    }
	    if (c==(gettext("u"))[0]) {
		*token = (pMS->get_csconv())[(*token)].clower;
	    }
	    if ((c==(gettext("u"))[0]) || (c==(gettext("i"))[0])) {
		struct wordlist* i = 
		    (struct wordlist *) malloc (sizeof(struct wordlist));
    		i->word = mystrdup(token);
		i->next = dicwords;
		dicwords = i;
		// save
		strcpy(buf,HOME);
#ifndef WINDOWS
		strcat(buf,"/");
#endif
		buf2 = buf+strlen(buf);
		strcat(buf,DICBASENAME);
		strcat(buf,basename(dicname,PATHSEP));
		if (save_privdic(buf2, buf, dicwords)) {
		    dicwords=NULL;
		} else {
		    fprintf(stderr,gettext("Cannot update personal dictionary."));
		    break;
		}
	    } // no break
	    if ((c==(gettext("u"))[0]) || (c==(gettext("i"))[0]) || (c==(gettext("a"))[0])) {
		modified=1;
		putdic(token,pMS);
		goto ki;
	    }
	    if (c==(gettext("o"))[0]) {
		modified=1;

		echo();
//		dialogscreen(token, filename, len, begin, wlst, ns, newline, rel, pre2, pre);
//		printw(gettext("Root word: "));
		printw(gettext("New root word: "));
		char w[BUFSIZ];
		getnstr(w,BUFSIZ);

		strcat(w,"//");
		char * w2 = w + strlen(w);
    		dialogscreen(token, filename, len, begin, wlst, ns, newline, rel, pre2, pre);
//		printw(gettext("Pattern: "));
		printw(gettext("Sample word: "));
		getnstr(w2,BUFSIZ);

		noecho();
				
///		fprintf(stderr,"X: %s\n",w); exit(1);
		if (putdic(w,pMS)) {

		    struct wordlist* i = 
			(struct wordlist *) malloc (sizeof(struct wordlist));
    		    i->word = mystrdup(w);
		    i->next = dicwords;
		    dicwords = i;
		    // save
		    
///		fprintf(stderr,"X: %s\n",i->word); endwin(); exit(1);
		    
		    strcpy(buf,HOME);
#ifndef WINDOWS
		    strcat(buf,"/");
#endif
		    buf2 = buf+strlen(buf);
		    strcat(buf,DICBASENAME);
		    strcat(buf,basename(dicname,PATHSEP));
		    if (save_privdic(buf2, buf, dicwords)) {
			dicwords=NULL;
		    } else {
			fprintf(stderr,gettext("Cannot update personal dictionary."));
			break;
		    }
		    
		} else {
    		    dialogscreen(token, filename, len, begin, wlst, ns, newline, rel, pre2, pre);
		    printw(gettext("Error: root word is not new, or sample word is not (accepted) root."));
		    break;
		}
		goto ki;
	    }
	    if (c==(gettext("x"))[0]) {
		return 1;
	    }
	    if (c==(gettext("q"))[0]) {
		if (modified) {
		    printw(gettext("Are you sure you want to throw away your changes? "));
		    if (getch()==(gettext("y"))[0]) return -1;
    		    dialogscreen(token, filename, len, begin, wlst, ns, newline, rel, pre2, pre);
		    break;		    
		} else {
		    return -1;
		}
	    }
	}
    }
    }
    ki: return 0;
}

int interactive_line(char * buf, MySpell * pMS, char * filename, FILE * tempfile,
    char ** ppre2, char ** ppre)
{
    char newline[BUFSIZ];
    int rel=0;
    int TeXMODE=1;
    int k;
    int dp;
    char * token;
    char * line=NULL;
    char prevnochar='\0';
    
    k = strlen(buf);
      
    // data line (not command)
    line = mystrdup(buf);
    token = buf;
    strcpy(newline,line);
    int dialogexit=0;

    for (int i=0; i<=k; i++) {
	if (!TESTAFF(wordchars, buf[i], wordchars_len)) {
	// first character of line or TeX definition
	    if ((buf+i-token==0) || ((TeXMODE==1) && (prevnochar=='\\'))) {
    		// previous non character
		prevnochar = buf[i];
	        token = buf+i+1;
		continue;
	    }
	    prevnochar = buf[i];
    	    buf[i] = '\0';
	    dp = pMS->spell(token);
	    if (!dp) {
		char ** wlst=NULL;
		int ns = pMS->suggest(&wlst,token);
		if (ns==0) {
		    dialogexit=dialog(pMS,token,filename,line,strlen(token),
				token-buf,wlst,ns,newline,&rel, *ppre2, *ppre);
		} else {	    
		    int nt = strlen(token);
		    while ((nt != 1) && (*(token+nt-1)=='.')) { // mistake...
			nt--;
			*(token+nt)='\0';
		    }
		    dialogexit=dialog(pMS,token,filename,line,nt,
				token-buf,wlst,ns,newline,&rel, *ppre2, *ppre);
		}
		for (int j=1; j < ns; j++) {
		    free(wlst[j]);
		}
		free(wlst);
	    }
	    if ((dialogexit==-1) || (dialogexit==1)) goto ki2;
	    if (dialogexit==2) {
		strcpy(buf,newline);
		free(line);
		line = mystrdup(buf);
		i = token - buf + rel - 1;
		k = strlen(buf);
		rel = 0;
		dialogexit = 0;
	    } else {
		token = buf+i+1; // lower limit
	    }
	}
    } // for

ki2: if (line) free(line);
    fprintf(tempfile,"%s\n",newline);
    if (*ppre2) free(*ppre2);
    *ppre2 = mystrdup(*ppre);
    if (*ppre) free(*ppre);
    *ppre = mystrdup(newline);
    return dialogexit;
} // interactive_line

void interactive_interface(MySpell * pMS, char * filename)
{
    char buf[BUFSIZ];
    FILE *text = fopen(filename,"r");
    int dialogexit;
    int check=1;

    char * pre2=NULL;
    char * pre=NULL;

    char tempname[] = "/tmp/mispell.XXXXXX";
    if (mkstemp(tempname) == -1) {
        fprintf(stderr,gettext("Can't create tempfile.\n"));
	endwin();
        exit(1);
    }

    FILE *tempfile = fopen(tempname,"w");

    if (text) {
	while(fgets(buf,BUFSIZ,text)) {
	    if (check) {
		*(buf + strlen(buf) - 1) = '\0';
		dialogexit=interactive_line(buf,pMS,filename,tempfile,&pre2,&pre);
		switch (dialogexit) {
		    case -1: {
			clear();
			refresh();
			unlink(tempname);
			endwin();
			exit(0);
		    }
		    case 1: {
			check = 0;
		    }
		}
	    } else {
		fprintf(tempfile,"%s",buf);
	    }
	}
	fclose(text);
	fclose(tempfile);

	if (pre) free(pre);
	if (pre2) free(pre2);

	// cat tempname >filename (keep file permission, 
	// and solve other filesystem problems, eg. different fs)
	if (! modified) {
	    unlink(tempname);
	} else {
	    strcpy(buf,"cat ");
	    strcat(buf,tempname);
	    strcat(buf," >");
	    strcat(buf,filename);
	    if (system(buf) == -1) {
    		fprintf(stderr,gettext("Can't move %s to %s.\n"),tempname,filename);
		endwin();
    		exit(1);
	    } else {
		unlink(tempname);
	    }
	}
    }
}

#endif
#endif

int main(int argc, char** argv)
{
MySpell * pMS = NULL;
int arg_files=-1; // first filename argumentum position

char buf[BUFSIZ];

int argstate=0;
dicname=NULL;

#ifdef CURSES
setlocale(LC_ALL, "");
textdomain("mispell");
#endif

log("START");


for(int i=1; i<argc; i++) {
    log(argv[i]);

    if (argstate==1) {
	if (dicname) free(dicname);
	dicname=mystrdup(argv[i]);
	argstate=0;
    }
    else if (strcmp(argv[i],"-d")==0) argstate=1;
    else if (strcmp(argv[i],"-h")==0) {
	fprintf(stderr,"MIspell ");
	fprintf(stderr,MISPELL_VERSION);
	fprintf(stderr,gettext(" - MySpell's Ispell Interface\n"));
	fprintf(stderr,gettext("Usage: mispell [-d dicname] [file(s)]\n"));
	fprintf(stderr,gettext("Example: mispell -d english file.txt\n"));
	fprintf(stderr,gettext("Author: La'szlo' Ne'meth, 2002, License: GNU GPL v2\n"));
	fprintf(stderr,gettext("E-mail: magyarispell@yahoogroups.com\n"));
	fprintf(stderr,gettext("MySpell-HU's author: Laszlo Nemeth (c) 2002\n"));
	fprintf(stderr,gettext("MySpell's author: Kevin Hendricks (c) 2001-2002\n"));
	fprintf(stderr,gettext("MySpell contributors: David Einstein, Laszlo Nemeth\n"));
	fprintf(stderr,gettext("Based on International Ispell's affix compression algorithm.\n"));
	fprintf(stderr,gettext("International Ispell's author: Geoff Kuenning\n"));
	exit(0);
    }
    else if ((strcmp(argv[i],"-vv")==0) || (strcmp(argv[i],"-v")==0)) {
	fprintf(stdout,gettext(MISPELL_HEADING));
	fprintf(stdout,MISPELL_VERSION);
	fprintf(stdout,"\n");
	exit(0);
    }
    else if ((arg_files==-1) && ((argv[i][0] != '-') && (argv[i][0] != '\0'))) {
	arg_files = i;
	if (! exist(argv[i])) { // first check (before time-consuming dic. load)
	    fprintf(stderr,gettext("Can't open %s.\n"),argv[i]);
#ifdef CURSES
	    endwin();
#endif
	    exit(1);
	}
    }
}

if (dicname==NULL) dicname = "default";

strcpy(buf,LIBDIR);
char * faff = buf+strlen(buf);
strcat(buf,dicname);
strcat(buf,".aff");
char * buf2 = buf+strlen(buf)+1;
strcpy(buf2,LIBDIR);
char * fdic = buf2+strlen(buf2);
strcat(buf2,dicname);
strcat(buf2,".dic");

if (exist(faff) && exist(fdic)) {
    pMS = new MySpell(faff,fdic);
}

if ((!pMS) && exist(buf) && exist(buf2)) {
    pMS = new MySpell(buf,buf2);
}

if (!pMS) {
    fprintf(stderr,gettext("Can't open affix or dictionary files.\n"));
    exit(1);
}

    /* open the private dictionaries */
    strcpy(buf,HOME);
#ifndef WINDOWS
    strcat(buf,"/");
#endif
    strcat(buf,DICBASENAME);
    strcat(buf,basename(dicname,PATHSEP));
    load_privdic(buf, pMS);
    strcpy(buf,DICBASENAME);
    strcat(buf,basename(dicname,PATHSEP));
    load_privdic(buf, pMS);

    wordchars = pMS->get_wordchars();
    if (!wordchars) {
	wordchars="qwertzuiopasdfghjklyxcvbnmQWERTZUIOPASDFGHJKLYXCVBNM";
    }
    ////fprintf(stderr,"WORDS: %s\n",wordchars);
    wordchars_len = strlen(wordchars);

    if (arg_files==-1) {
	    pipe_interface(pMS);
    } else {
#ifndef WINDOWS
#ifdef CURSES
	initscr();
	cbreak();
	noecho();
	nonl();
	intrflush(stdscr,FALSE);
	keypad(stdscr,TRUE);
	for (int i=arg_files; i<argc; i++) {
	    if (exist(argv[i])) {
		modified = 0;
	        interactive_interface(pMS,argv[i]);
	    } else {
	        fprintf(stderr,gettext("Can't open %s.\n"),argv[i]);
	        endwin();
	        exit(1);
	    }
	}
	clear();
	refresh();
	endwin();
#endif
#endif
    }

    // wordchars
    delete pMS;
    return 0;
}
